#include "extension.h"

#include "basehandle.h"
#include "vector.h"
#include "utlvector.h"
#include <shareddefs.h>
#include "igamemovement.h"
#include "usercmd.h"

SH_DECL_HOOK2_void(IGameMovement, ProcessMovement, SH_NOATTRIB, false, CBasePlayer *, CMoveData *);
SH_DECL_MANUALHOOK3_void(RunCommand, 0, 0, 0, CBasePlayer *, CUserCmd *, IMoveHelper *);

CmdFix g_CmdFix;		/**< Global singleton for extension's main interface */
SMEXT_LINK(&g_CmdFix);

ICvar *icvar = NULL;
IGameConfig *g_pGameConf = NULL;
IGameMovement *gameMovement = NULL;
void *playerMove = NULL;

ConVar cmdfix_kick("cmdfix_kick", "1", FCVAR_NONE, "Determines whether to kick a client when detected or to just block the cmd.");

bool CmdFix::SDK_OnLoad(char *error, size_t maxlength, bool late)
{
	if (!gameconfs->LoadGameConfigFile("cmdfix.games", &g_pGameConf, error, maxlength))
		return false;

	if (!g_pGameConf->GetAddress("PlayerMove", &playerMove) || !playerMove)
	{
		smutils->Format(error, maxlength, "Failed to find address for PlayerMove");
		return false;
	}

	int offset = 0;
	if (!g_pGameConf->GetOffset("CPlayerMove::RunCommand", &offset))
	{
		smutils->Format(error, maxlength, "Failed to find offset for CPlayerMove::RunCommand");
		return false;
	}
	SH_MANUALHOOK_RECONFIGURE(RunCommand, offset, 0, 0);

	gameconfs->CloseGameConfigFile(g_pGameConf);

	SH_ADD_HOOK(IGameMovement, ProcessMovement, gameMovement, SH_MEMBER(this, &CmdFix::Hook_ProcessMovement), false);
	SH_ADD_MANUALHOOK(RunCommand, playerMove, SH_MEMBER(this, &CmdFix::Hook_RunCommand), false);

	return true;
}

bool CmdFix::SDK_OnMetamodLoad(ISmmAPI *ismm, char *error, size_t maxlen, bool late)
{
	GET_V_IFACE_CURRENT(GetEngineFactory, icvar, ICvar, CVAR_INTERFACE_VERSION);
	GET_V_IFACE_CURRENT(GetServerFactory, gameMovement, IGameMovement, INTERFACENAME_GAMEMOVEMENT);

	g_pCVar = icvar;
	ConVar_Register(0, this);

	MathLib_Init(2.2f, 2.2f, 0.0f, 2);

	return true;
}

bool CmdFix::RegisterConCommandBase(ConCommandBase *pVar)
{
	return META_REGCVAR(pVar);
}

void CmdFix::SDK_OnUnload()
{
	SH_REMOVE_HOOK(IGameMovement, ProcessMovement, gameMovement, SH_MEMBER(this, &CmdFix::Hook_ProcessMovement), false);
	SH_REMOVE_MANUALHOOK(RunCommand, playerMove, SH_MEMBER(this, &CmdFix::Hook_RunCommand), false);
}

inline bool IsEntityCoordinateReasonable(const vec_t c)
{
	float r = 16384.0f; // k_flMaxEntityPosCoord
	return c > -r && c < r;
}

inline bool IsEntityQAngleReasonable(const QAngle &q)
{
	float r = 360.0 * 1000.0f; // k_flMaxEntityEulerAngle
	return
		q.x > -r && q.x < r &&
		q.y > -r && q.y < r &&
		q.z > -r && q.z < r;
}

bool IsUserCmdDataValid(CUserCmd *pCmd)
{
	return 
		// Prevent clients from sending invalid view angles to try to get leaf server code to crash
		(pCmd->viewangles.IsValid() && IsEntityQAngleReasonable(pCmd->viewangles)) &&
		// Movement ranges
		(IsFinite(pCmd->forwardmove) && IsEntityCoordinateReasonable(pCmd->forwardmove)) &&
		(IsFinite(pCmd->sidemove) && IsEntityCoordinateReasonable(pCmd->sidemove)) &&
		(IsFinite(pCmd->upmove) && IsEntityCoordinateReasonable(pCmd->upmove));
}

inline void MakeInert(CUserCmd *cmd)
{
	cmd->viewangles.x = 0.0f;
	cmd->viewangles.y = 0.0f;
	cmd->viewangles.z = 0.0f;
	cmd->forwardmove = 0.0f;
	cmd->sidemove = 0.0f;
	cmd->upmove = 0.0f;
	cmd->buttons = 0;
	cmd->impulse = 0;
}

void CmdFix::Hook_ProcessMovement(CBasePlayer *player, CMoveData *pMove)
{
	if (!player || !pMove)
		RETURN_META(MRES_IGNORED);

	IGamePlayer *pPlayer = playerhelpers->GetGamePlayer(gamehelpers->EntityToBCompatRef((CBaseEntity *)player));

	if (!pPlayer || pPlayer->IsFakeClient())
		RETURN_META(MRES_IGNORED);

	if (!pMove->m_vecViewAngles.IsValid() || !IsEntityQAngleReasonable(pMove->m_vecViewAngles) ||
		!pMove->m_vecAbsViewAngles.IsValid() || !IsEntityQAngleReasonable(pMove->m_vecAbsViewAngles))
		RETURN_META(MRES_SUPERCEDE);

	RETURN_META(MRES_IGNORED);
}

void CmdFix::Hook_RunCommand(CBasePlayer *player, CUserCmd *ucmd, IMoveHelper *moveHelper)
{
	if (!player || !moveHelper)
		RETURN_META(MRES_IGNORED);

	IGamePlayer *pPlayer = playerhelpers->GetGamePlayer(gamehelpers->EntityToBCompatRef((CBaseEntity *)player));

	if (!pPlayer || pPlayer->IsFakeClient())
		RETURN_META(MRES_IGNORED);

	// Validate values
	if (!IsUserCmdDataValid(ucmd))
	{
		MakeInert(ucmd);
		
		if (cmdfix_kick.GetBool() && !pPlayer->IsInKickQueue())
		{
			smutils->LogMessage(myself, "%s (ID: %s | IP: %s) was kicked for sending a corrupt user cmd.", pPlayer->GetName(), pPlayer->GetAuthString(false), pPlayer->GetIPAddress());
			gamehelpers->AddDelayedKick(pPlayer->GetIndex(), pPlayer->GetUserId(), "Corrupt user cmd");
		}
	}

	RETURN_META(MRES_IGNORED);
}
